#ifndef STOCHASTIC_HPP
#define STOCHASTIC_HPP

#include <climits>
#include <vector>
#include <string>
#include <unordered_map>
#include <unordered_set>
#include "polyomino.hpp"
#include "event.hpp"
#include "pugixml.hpp"

struct DimensionRestriction {
	Vec3 min, max;

	DimensionRestriction() {
		min = Vec3(INT_MIN, INT_MIN, INT_MIN);
		max = Vec3(INT_MAX, INT_MAX, INT_MAX);
	}
};

struct ShapeType
{
	std::vector<PolyominoType*> polyominoTypes;
	std::vector<Vec3> coords;

	Vec3 minCoord;
	Vec3 maxCoord;
	
	int id;
};

struct Shape
{
	ShapeType *type;
	Vec3 offset;

	Shape(ShapeType *type=nullptr, Vec3 offset=Vec3(0,0,0))
		: type(type), offset(offset) {}
};

inline bool operator==(const Shape &lhs, const Shape &rhs)
{
	return lhs.type == rhs.type and lhs.offset == rhs.offset;
}

namespace std
{
	template <>
	struct hash<Shape>
	{
		// FNV-1a hash implementation
		std::size_t operator()(const Shape &shape) const noexcept {
			const unsigned char* tptr = (const unsigned char*)&(shape.type);
			const unsigned char* xptr = (const unsigned char*)&(shape.offset.x);
			const unsigned char* yptr = (const unsigned char*)&(shape.offset.x);
			const unsigned char* zptr = (const unsigned char*)&(shape.offset.x);
			std::size_t h = FNV_OFFSET;
			for (unsigned i = 0; i < sizeof(ShapeType*); ++i)
				h = (h ^ tptr[i]) * FNV_PRIME;
			for (unsigned i = 0; i < sizeof(int); ++i)
				h = (h ^ xptr[i]) * FNV_PRIME;
			for (unsigned i = 0; i < sizeof(int); ++i)
				h = (h ^ yptr[i]) * FNV_PRIME;
			for (unsigned i = 0; i < sizeof(int); ++i)
				h = (h ^ zptr[i]) * FNV_PRIME;
			return h;
		}
	};
}

struct PrecomputedMaps
{
	std::vector<PolyominoType*> polyominoTypes;
	std::vector<ShapeType*> polyominoShapes;
	std::vector<ShapeType*> shapeTypes;
	std::vector<std::vector<Shape>> neighborLists;
	std::vector<std::vector<Shape>> overlapLists;
	std::vector<std::vector<std::unordered_map<Vec3, int>>> bindingMap;
};

class AbstractSimulation
{
	std::unordered_map<Shape, Polyomino> locationMap;
	std::unordered_set<Polyomino> frontier;
	std::vector<Polyomino> frontierList;
	std::unordered_map<Polyomino, int> frontierIds;

	// when frontier locations are added or removed, each grid location will be updated so that non-determinism can be evaluated
	std::unordered_map<Vec3, int> frontierGridLocationCounts;

	// the following may have entries for unoccupied shapes
	// keeps track of which shapes are overlapped
	std::unordered_map<Shape, int> overlappedMap;
	std::unordered_map<Polyomino, int> strengthMap;
	std::unordered_set<Polyomino> seedSet;

	pugi::xml_document *systemXMLDocument;

	std::mt19937 gen;

	int bindingThreshold;
	DimensionRestriction dimRestriction;
	int outputInterval;
	int numSteps;

	std::string outputDir;
	std::string outputName;

	PrecomputedMaps *sharedData;

	long long rngSeed;
	int threadId;
	int steps;

	bool reportNonDeterminism;

	void addFrontier(Polyomino polyomino);
	void removeFrontier(Polyomino polyomino);

	void checkNonDeterminism();

	void initSystemXML();
	void writeSystemXML();

public:
	AbstractSimulation(PrecomputedMaps *sharedData,
		const std::unordered_set<Polyomino> &seedPolyominoes,
		const std::unordered_set<Polyomino> &assemblyPolyominoes,
		int bindingThreshold,
		DimensionRestriction dimRestrictions,
		int outputInterval, std::string outputDir, std::string OutputName, int numSteps, int threadId, bool reportNonDeterminism, long long rngSeed);

	// AbstractSimulation(const AbstractSimulation &sim);

	bool step();
	void addPolyomino(Polyomino polyomino, bool seed=false);

	void operator()();
};

class KineticSimulation
{
	// keeps track of attached polyominoes by shape
	// should only have entries for shapes with attached polyominoes
	//std::unordered_map<Shape, State> shapeMap;
	std::unordered_map<Shape, Polyomino> locationMap;
	
	// the following may have entries for unoccupied shapes
	// keeps track of which shapes are overlapped
	std::unordered_map<Shape, int> overlappedMap;
	std::unordered_map<Polyomino, int> strengthMap;
	std::unordered_set<Polyomino> seedSet;

	double Gmc, Gse, kf;
	StochasticEventSet *eventSet;

	DimensionRestriction dimRestriction;
	int minBinding;
	int outputInterval;
	int numSteps;
	int maxAttachments;

	std::string outputDir;
	std::string outputName;

	PrecomputedMaps *sharedData;

	long long rngSeed;
	int threadId;

	double time;
	int steps;

	pugi::xml_document *systemXMLDocument;

	void seedRng(int seed);

	void initSystemXML();
	void writeSystemXML(std::string note="");

public:

	KineticSimulation(PrecomputedMaps *sharedData,
		const std::unordered_set<Polyomino> &seedPolyominoes,
		const std::unordered_set<Polyomino> &assemblyPolyominoes, 
		double Gmc, double Gse, double kf,
		DimensionRestriction dimRestrictions, int minBinding, 
		int outputInterval, std::string outputDir, std::string outputName, int numSteps, int maxAttachments, int threadId, long long rngSeed);

	void step();
	void addPolyomino(Polyomino polyomino, bool seed=false);
	void removePolyomino(Polyomino polyomino);

	void operator()();
};

struct StochasticSim
{
	// keeps track of attached polyominoes by shape
	// should only have entries for shapes with attached polyominoes
	//std::unordered_map<Shape, State> shapeMap;
	std::unordered_map<Shape, Polyomino> locationMap;
	
	// the following may have entries for unoccupied shapes
	// keeps track of which shapes are overlapped
	std::unordered_map<Shape, int> overlappedMap;
	std::unordered_map<Polyomino, int> strengthMap;
	std::unordered_set<Polyomino> seedSet;

	// precomputation maps
	std::vector<PolyominoType*> *polyominoTypes;
	std::vector<ShapeType*> *polyominoShapes;
	std::vector<ShapeType*> *shapeTypes;

	// given the id of a shapeType, stores the relative neighbors and overlaps
	std::vector<std::vector<Shape>> *neighborLists;
	std::vector<std::vector<Shape>> *overlapLists;

	std::vector<std::vector<std::unordered_map<Vec3, int>>> *bindingMap;

	double Gmc, Gse, kf;
	StochasticEventSet *eventSet;

	DimensionRestriction dimRestriction;
	int minBinding;

	double time;
	int steps;

	StochasticSim(std::vector<PolyominoType*> *types, double Gmc, double Gse, double kf,
		      DimensionRestriction dimRestriction, int minBinding);

	void setPrecomputationPointers(std::vector<ShapeType*> *polyominoShapes,
								   std::vector<ShapeType*> *shapeTypes,
								   std::vector<std::vector<Shape>> *neighborLists,
								   std::vector<std::vector<Shape>> *overlapLists,
								   std::vector<std::vector<std::unordered_map<Vec3, int>>> *bindingMap);

	void step();
	void addPolyomino(const Polyomino &polyomino, bool seed=false);
	void removePolyomino(const Polyomino &polyomino);
private:
	// void computeShapes();
	// void computeBoundingBoxes();
	// void checkNeighbors(ShapeType *st1, ShapeType *st2);
	// void computeNeighborLists();
	// void computeBindingMap();
};

#endif
