#include <iostream>
#include <algorithm>
#include <chrono>
#include "stochastic.hpp"

AbstractSimulation::AbstractSimulation(PrecomputedMaps *sharedData,
	const std::unordered_set<Polyomino> &seedPolyominoes,
	const std::unordered_set<Polyomino> &assemblyPolyominoes, 
	int bindingThreshold,
	DimensionRestriction dimRestrictions, 
	int outputInterval, std::string outputDir, std::string outputName, int numSteps, int threadId, bool reportNonDeterminism,
	long long rngSeed)
	: bindingThreshold(bindingThreshold), dimRestriction(dimRestrictions), outputInterval(outputInterval), 
	outputDir(outputDir), outputName(outputName), numSteps(numSteps), threadId(threadId), reportNonDeterminism(reportNonDeterminism), rngSeed(rngSeed)
{
	this->sharedData = sharedData;

	this->steps = 0;

	if (this->rngSeed == -1 or this->threadId != 0) {
		this->rngSeed = std::chrono::system_clock::now().time_since_epoch().count();
	}

	this->gen.seed(this->rngSeed);
	std::cout << "TID " << this->threadId << " using RNG seed: " << this->rngSeed << std::endl;

	for (Polyomino poly : seedPolyominoes)
		this->addPolyomino(poly, true);

	for (Polyomino poly : assemblyPolyominoes)
		this->addPolyomino(poly, false);

	this->initSystemXML();
}

void AbstractSimulation::addFrontier(Polyomino polyomino)
{
	if (!frontier.count(polyomino))
	{
		frontierIds.emplace(polyomino, frontierList.size());
		frontier.insert(polyomino).first;
		// Polyomino frontierPoly = *(frontier.find(polyomino));

		frontierList.push_back(polyomino);

		// add each grid location of the polyomino to our count
		for (Block block : polyomino.type->blocks) {
			Vec3 coord = block.coord + polyomino.offset;

			if (this->frontierGridLocationCounts.count(coord) == 0)
				this->frontierGridLocationCounts[coord] = 0;

			this->frontierGridLocationCounts[coord] += 1;
		}
	}
	else
	{
		std::cout << "Attempting to add existing frontier poly" << std::endl;
	}
}

void AbstractSimulation::removeFrontier(Polyomino polyomino)
{
	if (frontier.count(polyomino))
	{
		int id = frontierIds[polyomino];

		if (id + 1 != frontierList.size())
		{
			Polyomino curr = polyomino;

			int lastId = frontierList.size() - 1;
			Polyomino last = frontierList[lastId];

			frontierList[lastId] = curr;
			frontierList[id] = last;

			frontierIds[last] = id;
		}

		frontier.erase(polyomino);
		frontierIds.erase(polyomino);
		frontierList.pop_back();

		for (Block block : polyomino.type->blocks) {
			Vec3 coord = block.coord + polyomino.offset;

			int &count = this->frontierGridLocationCounts[coord];
			count -= 1;
			if (count <= 0) {
				this->frontierGridLocationCounts.erase(coord);
			}
		}
	}
	else
	{
		std::cout << "Attempting to remove non-existing frontier poly" << std::endl;
	}
}

void AbstractSimulation::checkNonDeterminism() {
	for (auto it = this->frontierGridLocationCounts.begin(); it != this->frontierGridLocationCounts.end(); ++it) {
		Vec3 coord = it->first;
		int count = it->second;

		// non-determinism found
		if (count > 1) {
			std::cout << "TID " << this->threadId << ": Non-Determinism detected at coordinates " << coord.toString() << std::endl;
		}
	}
}

void AbstractSimulation::addPolyomino(Polyomino polyomino, bool seed)
{
	int id1 = polyomino.type->id;
	
	// update neighbors
	ShapeType *sType = this->sharedData->polyominoShapes[id1];
	Shape shape(sType, polyomino.offset);
	
	if (locationMap.count(shape)) {
		std::string errString = "attempting to add polyomino \"" +
			polyomino.type->name + "\" in occupied location " +
			polyomino.offset.toString();
		
		throw std::runtime_error(errString);
	}

	int strength = strengthMap[polyomino];

	polyomino.stepAdded = this->steps;

	polyomino.initialBindStrength = strength;
	polyomino.currentBindStrength = strength;

	locationMap.emplace(shape, polyomino);
	if (seed) seedSet.insert(polyomino);
	// else eventSet->addDetachmentEvent(polyomino, strength);

	for (Shape nbr : this->sharedData->neighborLists[sType->id]) {

		Vec3 offset = nbr.offset;
		nbr.offset += polyomino.offset;

		// ignore neighbors that don't satisfy dim restrictions
		if (nbr.offset < dimRestriction.min or nbr.offset >= dimRestriction.max)
			continue;

		// polyomino already exists here, skip
		if (locationMap.count(nbr))
			continue;

		for (PolyominoType *pType : nbr.type->polyominoTypes)
		{
			Polyomino polyominoNbr(pType, nbr.offset);

			int id2 = pType->id;
			int relStrength = this->sharedData->bindingMap[id1][id2][offset];

			int &strengthNbr = strengthMap[polyominoNbr];
			int oldStrengthNbr = strengthNbr;
			strengthNbr += relStrength;

			if (overlappedMap[nbr] == 0 and
				oldStrengthNbr < this->bindingThreshold and 
				strengthNbr >= this->bindingThreshold)
			{
				addFrontier(polyominoNbr);
				// frontier.insert(polyominoNbr);
			}
		}
	}

	// remove overlapping attachment events
	for (Shape ovr : this->sharedData->overlapLists[sType->id]) {
		ovr.offset += polyomino.offset;

		if (ovr.offset < dimRestriction.min or ovr.offset >= dimRestriction.max)
			continue;
		
		int overlappedOvr = ++(overlappedMap[ovr]);

		for (PolyominoType *pType : ovr.type->polyominoTypes) {
			Polyomino polyominoOvr(pType, ovr.offset);
			int strengthOvr = strengthMap[polyominoOvr];
			if (strengthOvr >= this->bindingThreshold and overlappedOvr == 1) {
				// frontier.erase(polyominoOvr);
				removeFrontier(polyominoOvr);
			}
		}
	}
}

bool AbstractSimulation::step()
{
	if (this->reportNonDeterminism)
		this->checkNonDeterminism();

	int size = this->frontierList.size();
	if (size == 0)
		return false;

	auto uniformInt = std::uniform_int_distribution<int>(0, size-1);

	Polyomino polyomino = this->frontierList[uniformInt(gen)];
	// this->removeFrontier(polyomino);
	this->addPolyomino(polyomino);
	this->steps += 1;

	return true;
}

void AbstractSimulation::operator()()
{
	for (int i = 1; i <= this->numSteps; i++)
	{
		if(!this->step())
		{
			std::cout << "THREAD " << this->threadId << ": ";
			std::cout << "empty frontier, step " << i << "/" << this->numSteps << " (" << this->locationMap.size() << " tiles)" << std::endl;
			this->writeSystemXML();
			break;
		}
		if (i % this->outputInterval == 0)
		{
			std::cout << "THREAD " << this->threadId << ": ";
			std::cout << "step " << i << "/" << this->numSteps << " (" << this->locationMap.size() << " tiles)" << std::endl;
			this->writeSystemXML();
		}
	}
}

void AbstractSimulation::initSystemXML()
{
	this->systemXMLDocument = new pugi::xml_document();

	pugi::xml_node sysNode = this->systemXMLDocument->append_child("PolyominoSystem");

	// todo
	// pugi::xml_node dimRestrictionsNode = sysNode.append_child("dim_restrictions");
	// dimRestrictionsNode.append_child(pugi::node_pcdata).set_value("");

	pugi::xml_node rotationNode = sysNode.append_child("rotation");
	rotationNode.append_child(pugi::node_pcdata).set_value("False");

	pugi::xml_node minBindingNode = sysNode.append_child("bindingThreshold");
	minBindingNode.append_child(pugi::node_pcdata).set_value(std::to_string(this->bindingThreshold).c_str());
	
	pugi::xml_node stepsNode = sysNode.append_child("steps");
	stepsNode.append_child(pugi::node_pcdata).set_value(std::to_string(this->steps).c_str());

	pugi::xml_node typesNode = sysNode.append_child("PolyominoTypes");
	for (const PolyominoType *type : this->sharedData->polyominoTypes)
	{
		pugi::xml_node polyominoNode = typesNode.append_child("PolyominoType");
		
		pugi::xml_node nameNode = polyominoNode.append_child("name");
		nameNode.append_child(pugi::node_pcdata).set_value(type->name.c_str());

		pugi::xml_node colorNode = polyominoNode.append_child("color");
		colorNode.append_child(pugi::node_pcdata).set_value(type->color.c_str());

		pugi::xml_node concNode = polyominoNode.append_child("concentration");
		concNode.append_child(pugi::node_pcdata).set_value(std::to_string(type->concentration).c_str());

		pugi::xml_node blocksNode = polyominoNode.append_child("blocks");
		for (const Block &block : type->blocks)
		{
			pugi::xml_node blockNode = blocksNode.append_child("block");

			pugi::xml_node coordsNode = blockNode.append_child("coords");
			coordsNode.append_child(pugi::node_pcdata).set_value(block.coord.toString().c_str());

			pugi::xml_node domainsNode = blockNode.append_child("domains");
			for (const Domain &domain : block.domains)
			{
				pugi::xml_node domainNode = domainsNode.append_child("domain");

				pugi::xml_node labelNode = domainNode.append_child("label");
				labelNode.append_child(pugi::node_pcdata).set_value(domain.label.c_str());

				pugi::xml_node dirNode = domainNode.append_child("direction");
				char dirStr[2] = " ";
				switch(domain.direction)
				{
					case NORTH: dirStr[0] = 'N'; break;
					case SOUTH: dirStr[0] = 'S'; break;
					case EAST:  dirStr[0] = 'E'; break;
					case WEST:  dirStr[0] = 'W'; break;
					case UP:    dirStr[0] = 'U'; break;
					case DOWN:  dirStr[0] = 'D'; break;
				}
				dirNode.append_child(pugi::node_pcdata).set_value(dirStr);

				pugi::xml_node strengthNode = domainNode.append_child("strength");
				strengthNode.append_child(pugi::node_pcdata).set_value(std::to_string(domain.strength).c_str());
			}
		}
	}

	pugi::xml_node seedNode = sysNode.append_child("seed");

	pugi::xml_node seedPolysNode = seedNode.append_child("Polyominoes");
	for (const Polyomino &poly : this->seedSet)
	{
		pugi::xml_node polyNode = seedPolysNode.append_child("Polyomino");
		pugi::xml_node typeNode = polyNode.append_child("PolyominoType");
		typeNode.append_child(pugi::node_pcdata).set_value(poly.type->name.c_str());

		pugi::xml_node transNode = polyNode.append_child("translation");
		transNode.append_child(pugi::node_pcdata).set_value(poly.offset.toString().c_str());
	}

	pugi::xml_node assemblyNode = sysNode.append_child("assembly");

	pugi::xml_node assemblyPolysNode = assemblyNode.append_child("Polyominoes");
	for (const auto &kv : this->locationMap)
	{
		const Polyomino &poly = kv.second;
		pugi::xml_node polyNode = assemblyPolysNode.append_child("Polyomino");
		pugi::xml_node typeNode = polyNode.append_child("PolyominoType");
		typeNode.append_child(pugi::node_pcdata).set_value(poly.type->name.c_str());

		pugi::xml_node transNode = polyNode.append_child("translation");
		transNode.append_child(pugi::node_pcdata).set_value(poly.offset.toString().c_str());
	}


	// sysNode.remove_child(systemXMLDocument.child("assembly"));
	// pugi::xml_node assemblyNode = sysNode.append_child("assembly");
	// pugi::xml_node polyominoesNode = assemblyNode.append_child("Polyominoes");

	// for (auto kv : sim->locationMap) {
	// 	Polyomino p = kv.second;

	// 	pugi::xml_node polyominoNode = polyominoesNode.append_child("Polyomino");
	// 	pugi::xml_node typeNode = polyominoNode.append_child("PolyominoType");
	// 	typeNode.append_child(pugi::node_pcdata).set_value(p.type->name.c_str());
	// 	pugi::xml_node translationNode = polyominoNode.append_child("translation");
	// 	translationNode.append_child(pugi::node_pcdata).set_value(p.offset.toString().c_str());
	// }
}

void AbstractSimulation::writeSystemXML()
{
	pugi::xml_node sysNode = this->systemXMLDocument->child("PolyominoSystem");

	sysNode.remove_child("steps");
	pugi::xml_node stepsNode = sysNode.prepend_child("steps");
	stepsNode.append_child(pugi::node_pcdata).set_value(std::to_string(this->steps).c_str());

	sysNode.remove_child("assembly");
	pugi::xml_node assemblyNode = sysNode.append_child("assembly");

	pugi::xml_node assemblyPolysNode = assemblyNode.append_child("Polyominoes");
	for (const auto &kv : this->locationMap)
	{
		const Polyomino &poly = kv.second;
		pugi::xml_node polyNode = assemblyPolysNode.append_child("Polyomino");
		pugi::xml_node typeNode = polyNode.append_child("PolyominoType");
		typeNode.append_child(pugi::node_pcdata).set_value(poly.type->name.c_str());

		pugi::xml_node transNode = polyNode.append_child("translation");
		transNode.append_child(pugi::node_pcdata).set_value(poly.offset.toString().c_str());

		pugi::xml_node polyStepNode = polyNode.append_child("stepAdded");
		polyStepNode.append_child(pugi::node_pcdata).set_value(std::to_string(poly.stepAdded).c_str());

		// pugi::xml_node polyTimeNode = polyNode.append_child("timeAdded");
		// polyTimeNode.append_child(pugi::node_pcdata).set_value(std::to_string(poly.timeAdded).c_str());

		pugi::xml_node polyBindNode = polyNode.append_child("bindStrength");
		polyBindNode.append_child(pugi::node_pcdata).set_value(std::to_string(poly.currentBindStrength).c_str());

		// pugi::xml_node polyInitialBindNode = polyNode.append_child("initialBindStrength");
		// polyInitialBindNode.append_child(pugi::node_pcdata).set_value(std::to_string(poly.initialBindStrength).c_str());
	}
	// for (const auto &kv : this->locationMap)
	// {
	// 	const Polyomino &poly = kv.second;
	// 	pugi::xml_node polyNode = assemblyPolysNode.append_child("Polyomino");
	// 	pugi::xml_node typeNode = polyNode.append_child("PolyominoType");
	// 	typeNode.append_child(pugi::node_pcdata).set_value(poly.type->name.c_str());

	// 	pugi::xml_node transNode = polyNode.append_child("translation");
	// 	transNode.append_child(pugi::node_pcdata).set_value(poly.offset.toString().c_str());
	// }

	std::string filename = this->outputDir + this->outputName;
	filename += "_tid" + std::to_string(this->threadId);
	filename += "_step" + std::to_string(this->steps);
	filename += ".xml";

	this->systemXMLDocument->save_file(filename.c_str());

}

KineticSimulation::KineticSimulation(PrecomputedMaps *sharedData,
	const std::unordered_set<Polyomino> &seedPolyominoes,
	const std::unordered_set<Polyomino> &assemblyPolyominoes, 
	double Gmc, double Gse, double kf,
	DimensionRestriction dimRestrictions, int minBinding, 
	int outputInterval, std::string outputDir, std::string outputName, int numSteps, int maxAttachments, int threadId, long long rngSeed)
	: Gmc(Gmc), Gse(Gse), kf(kf), dimRestriction(dimRestrictions), minBinding(minBinding), outputInterval(outputInterval), 
	outputDir(outputDir), outputName(outputName), numSteps(numSteps), maxAttachments(maxAttachments), threadId(threadId), rngSeed(rngSeed)
{
	this->sharedData = sharedData;

	int maxStrength = 0;
	for (PolyominoType *type : sharedData->polyominoTypes)
	{
		// std::cout << type->name << " " << std::endl;
		int strength = 0;
		for (const Block &block : type->blocks)
			for (const Domain &domain : block.domains)
				strength += domain.strength;

		if (maxStrength < strength)
			maxStrength = strength;
	}

	if (this->rngSeed == -1 or this->threadId != 0) {
		this->rngSeed = std::chrono::system_clock::now().time_since_epoch().count();
	}

	this->eventSet = new StochasticEventSet(sharedData->polyominoTypes, maxStrength, Gmc, Gse, kf);
	this->eventSet->gen.seed(this->rngSeed);

	std::cout << "TID " << this->threadId << " using RNG seed: " << this->rngSeed << std::endl;

	this->time = 0.0;
	this->steps = 0;

	for (Polyomino poly : seedPolyominoes)
		this->addPolyomino(poly, true);

	for (Polyomino poly : assemblyPolyominoes)
		this->addPolyomino(poly, false);

	this->initSystemXML();
}

void KineticSimulation::step()
{
	StochasticEvent event = eventSet->sample();
	
	if (event.type == ATTACHMENT) {
		addPolyomino(event.polyomino);
	} else {
		removePolyomino(event.polyomino);
	}

	this->time += event.dt;
	this->steps += 1;
}

void KineticSimulation::operator()()
{
	std::string note = "";
	int i;

	for (i = 1; i <= this->numSteps; i++)
	{
		this->step();

		if (this->maxAttachments > 0) {
			if (this->locationMap.size() - this->seedSet.size() >= this->maxAttachments) {
				std::cout << "THREAD " << this->threadId << ": ";
				std::cout << "MAX ATTACHMENTS REACHED: step " << i << "/" << this->numSteps << " (" << this->locationMap.size() << " tiles)" << std::endl;
				note = "MAX ATTACHMENTS REACHED";
				break;
			}
		}

		if (i % this->outputInterval == 0)
		{
			std::cout << "THREAD " << this->threadId << ": ";
			std::cout << "step " << i << "/" << this->numSteps << " (" << this->locationMap.size() << " tiles)" << std::endl;
			
			if (i != this->numSteps) {
				this->writeSystemXML();
			}
		}
	}

	this->writeSystemXML(note);
}

void KineticSimulation::initSystemXML()
{
	this->systemXMLDocument = new pugi::xml_document();

	pugi::xml_node sysNode = this->systemXMLDocument->append_child("PolyominoSystem");

	pugi::xml_node GmcNode = sysNode.append_child("Gmc");
	GmcNode.append_child(pugi::node_pcdata).set_value(std::to_string(this->Gmc).c_str());

	pugi::xml_node GseNode = sysNode.append_child("Gse");
	GseNode.append_child(pugi::node_pcdata).set_value(std::to_string(this->Gse).c_str());

	pugi::xml_node kfNode = sysNode.append_child("forwardRate");
	kfNode.append_child(pugi::node_pcdata).set_value(std::to_string(this->kf).c_str());

	// todo
	// pugi::xml_node dimRestrictionsNode = sysNode.append_child("dim_restrictions");
	// dimRestrictionsNode.append_child(pugi::node_pcdata).set_value("");

	pugi::xml_node rotationNode = sysNode.append_child("rotation");
	rotationNode.append_child(pugi::node_pcdata).set_value("False");

	pugi::xml_node minBindingNode = sysNode.append_child("min_binding");
	minBindingNode.append_child(pugi::node_pcdata).set_value(std::to_string(this->minBinding).c_str());

	pugi::xml_node maxAttachmentsNode = sysNode.append_child("max_attachments");
	maxAttachmentsNode.append_child(pugi::node_pcdata).set_value(std::to_string(this->maxAttachments).c_str());

	pugi::xml_node timeNode = sysNode.append_child("time");
	timeNode.append_child(pugi::node_pcdata).set_value(std::to_string(this->time).c_str());
	// timeNode.append_child(pugi::node_pcdata).set_value(this->time);
	
	pugi::xml_node stepsNode = sysNode.append_child("steps");
	stepsNode.append_child(pugi::node_pcdata).set_value(std::to_string(this->steps).c_str());

	pugi::xml_node typesNode = sysNode.append_child("PolyominoTypes");
	for (const PolyominoType *type : this->sharedData->polyominoTypes)
	{
		pugi::xml_node polyominoNode = typesNode.append_child("PolyominoType");
		
		pugi::xml_node nameNode = polyominoNode.append_child("name");
		nameNode.append_child(pugi::node_pcdata).set_value(type->name.c_str());

		pugi::xml_node colorNode = polyominoNode.append_child("color");
		colorNode.append_child(pugi::node_pcdata).set_value(type->color.c_str());

		pugi::xml_node concNode = polyominoNode.append_child("concentration");
		concNode.append_child(pugi::node_pcdata).set_value(std::to_string(type->concentration).c_str());

		pugi::xml_node blocksNode = polyominoNode.append_child("blocks");
		for (const Block &block : type->blocks)
		{
			pugi::xml_node blockNode = blocksNode.append_child("block");

			pugi::xml_node coordsNode = blockNode.append_child("coords");
			coordsNode.append_child(pugi::node_pcdata).set_value(block.coord.toString().c_str());

			pugi::xml_node domainsNode = blockNode.append_child("domains");
			for (const Domain &domain : block.domains)
			{
				pugi::xml_node domainNode = domainsNode.append_child("domain");

				pugi::xml_node labelNode = domainNode.append_child("label");
				labelNode.append_child(pugi::node_pcdata).set_value(domain.label.c_str());

				pugi::xml_node dirNode = domainNode.append_child("direction");
				char dirStr[2] = " ";
				switch(domain.direction)
				{
					case NORTH: dirStr[0] = 'N'; break;
					case SOUTH: dirStr[0] = 'S'; break;
					case EAST:  dirStr[0] = 'E'; break;
					case WEST:  dirStr[0] = 'W'; break;
					case UP:    dirStr[0] = 'U'; break;
					case DOWN:  dirStr[0] = 'D'; break;
				}
				dirNode.append_child(pugi::node_pcdata).set_value(dirStr);

				pugi::xml_node strengthNode = domainNode.append_child("strength");
				strengthNode.append_child(pugi::node_pcdata).set_value(std::to_string(domain.strength).c_str());
			}
		}
	}

	pugi::xml_node seedNode = sysNode.append_child("seed");

	pugi::xml_node seedPolysNode = seedNode.append_child("Polyominoes");
	for (const Polyomino &poly : this->seedSet)
	{
		pugi::xml_node polyNode = seedPolysNode.append_child("Polyomino");
		pugi::xml_node typeNode = polyNode.append_child("PolyominoType");
		typeNode.append_child(pugi::node_pcdata).set_value(poly.type->name.c_str());

		pugi::xml_node transNode = polyNode.append_child("translation");
		transNode.append_child(pugi::node_pcdata).set_value(poly.offset.toString().c_str());
	}

	pugi::xml_node assemblyNode = sysNode.append_child("assembly");

	pugi::xml_node assemblyPolysNode = assemblyNode.append_child("Polyominoes");
	for (const auto &kv : this->locationMap)
	{
		const Polyomino &poly = kv.second;
		pugi::xml_node polyNode = assemblyPolysNode.append_child("Polyomino");
		pugi::xml_node typeNode = polyNode.append_child("PolyominoType");
		typeNode.append_child(pugi::node_pcdata).set_value(poly.type->name.c_str());

		pugi::xml_node transNode = polyNode.append_child("translation");
		transNode.append_child(pugi::node_pcdata).set_value(poly.offset.toString().c_str());

		pugi::xml_node polyStepNode = polyNode.append_child("stepAdded");
		polyStepNode.append_child(pugi::node_pcdata).set_value(std::to_string(poly.stepAdded).c_str());

		pugi::xml_node polyTimeNode = polyNode.append_child("timeAdded");
		polyTimeNode.append_child(pugi::node_pcdata).set_value(std::to_string(poly.timeAdded).c_str());

		pugi::xml_node polyBindNode = polyNode.append_child("bindStrength");
		polyBindNode.append_child(pugi::node_pcdata).set_value(std::to_string(poly.currentBindStrength).c_str());

		pugi::xml_node polyInitialBindNode = polyNode.append_child("initialBindStrength");
		polyInitialBindNode.append_child(pugi::node_pcdata).set_value(std::to_string(poly.initialBindStrength).c_str());
	}


	// sysNode.remove_child(systemXMLDocument.child("assembly"));
	// pugi::xml_node assemblyNode = sysNode.append_child("assembly");
	// pugi::xml_node polyominoesNode = assemblyNode.append_child("Polyominoes");

	// for (auto kv : sim->locationMap) {
	// 	Polyomino p = kv.second;

	// 	pugi::xml_node polyominoNode = polyominoesNode.append_child("Polyomino");
	// 	pugi::xml_node typeNode = polyominoNode.append_child("PolyominoType");
	// 	typeNode.append_child(pugi::node_pcdata).set_value(p.type->name.c_str());
	// 	pugi::xml_node translationNode = polyominoNode.append_child("translation");
	// 	translationNode.append_child(pugi::node_pcdata).set_value(p.offset.toString().c_str());
	// }
}

void KineticSimulation::writeSystemXML(std::string note)
{
	pugi::xml_node sysNode = this->systemXMLDocument->child("PolyominoSystem");

	sysNode.remove_child("time");
	pugi::xml_node timeNode = sysNode.prepend_child("time");
	timeNode.append_child(pugi::node_pcdata).set_value(std::to_string(this->time).c_str());
	// timeNode.append_child(pugi::node_pcdata).set_value(this->time);
	
	sysNode.remove_child("steps");
	pugi::xml_node stepsNode = sysNode.prepend_child("steps");
	stepsNode.append_child(pugi::node_pcdata).set_value(std::to_string(this->steps).c_str());

	sysNode.remove_child("assembly");
	pugi::xml_node assemblyNode = sysNode.append_child("assembly");

	sysNode.remove_child("note");
	pugi::xml_node noteNode = sysNode.prepend_child("note");
	noteNode.append_child(pugi::node_pcdata).set_value(note.c_str());

	pugi::xml_node assemblyPolysNode = assemblyNode.append_child("Polyominoes");
	for (const auto &kv : this->locationMap)
	{
		const Polyomino &poly = kv.second;
		pugi::xml_node polyNode = assemblyPolysNode.append_child("Polyomino");
		pugi::xml_node typeNode = polyNode.append_child("PolyominoType");
		typeNode.append_child(pugi::node_pcdata).set_value(poly.type->name.c_str());

		pugi::xml_node transNode = polyNode.append_child("translation");
		transNode.append_child(pugi::node_pcdata).set_value(poly.offset.toString().c_str());

		pugi::xml_node polyStepNode = polyNode.append_child("stepAdded");
		polyStepNode.append_child(pugi::node_pcdata).set_value(std::to_string(poly.stepAdded).c_str());

		pugi::xml_node polyTimeNode = polyNode.append_child("timeAdded");
		polyTimeNode.append_child(pugi::node_pcdata).set_value(std::to_string(poly.timeAdded).c_str());

		pugi::xml_node polyBindNode = polyNode.append_child("bindStrength");
		polyBindNode.append_child(pugi::node_pcdata).set_value(std::to_string(poly.currentBindStrength).c_str());

		pugi::xml_node polyInitialBindNode = polyNode.append_child("initialBindStrength");
		polyInitialBindNode.append_child(pugi::node_pcdata).set_value(std::to_string(poly.initialBindStrength).c_str());
	}

	std::string filename = this->outputDir + this->outputName;
	filename += "_tid" + std::to_string(this->threadId);
	filename += "_step" + std::to_string(this->steps);
	filename += ".xml";

	this->systemXMLDocument->save_file(filename.c_str());

}

void KineticSimulation::addPolyomino(Polyomino polyomino, bool seed)
{
	// std::cout << "Attempting to add polyomino " << polyomino.type->name << std::endl;
	// std::cout << "seed = " << seed << std::endl;

	int id1 = polyomino.type->id;
	
	// update neighbors
	ShapeType *sType = this->sharedData->polyominoShapes[id1];
	Shape shape(sType, polyomino.offset);
	
	if (locationMap.count(shape)) {
		std::string errString = "attempting to add polyomino \"" +
			polyomino.type->name + "\" in occupied location " +
			polyomino.offset.toString();
		
		throw std::runtime_error(errString);
	}

	int strength = strengthMap[polyomino];

	polyomino.stepAdded = this->steps;
	polyomino.timeAdded = this->time;

	polyomino.initialBindStrength = strength;
	polyomino.currentBindStrength = strength;

	locationMap.emplace(shape, polyomino);
	if (seed) seedSet.insert(polyomino);
	else eventSet->addDetachmentEvent(polyomino, strength);
	
	for (Shape nbr : this->sharedData->neighborLists[sType->id]) {
		
		Vec3 offset = nbr.offset;
		nbr.offset += polyomino.offset;

		// std::cout << nbr.offset.toString() << " found from poly " << polyomino.offset.toString() << std::endl;

		// ignore neighbors that don't satisfy dim restrictions
		if (nbr.offset < dimRestriction.min or nbr.offset >= dimRestriction.max)
			continue;

		// std::cout << dimRestriction.min.toString() << " " <<  dimRestriction.max.toString() << std::endl;
		// if (nbr.offset == Vec3(0,4,0))
		// {
		// 	for (auto *pType : nbr.type->polyominoTypes)
		// 		std::cout << pType->name << std::endl;
		// 	// for (Vec3 coord : nbr.type->coords)
		// 	// 	std::cout << coord.toString() << " " << std::endl;
		// 	// std::cout << std::endl;
		// }

		Polyomino existing = polyomino;
		if (locationMap.count(nbr))
			existing = locationMap[nbr];
		
		for (PolyominoType *pType : nbr.type->polyominoTypes) {

			Polyomino polyominoNbr(pType, nbr.offset);

			int id2 = pType->id;
			int relStrength = this->sharedData->bindingMap[id1][id2][offset];

			int &strengthNbr = strengthMap[polyominoNbr];
			int oldStrengthNbr = strengthNbr;
			strengthNbr += relStrength;

			if (polyominoNbr == existing and seedSet.count(polyominoNbr) == 0) {
				locationMap[nbr].currentBindStrength = strengthNbr;

				// std::cout << "updating attachment event for neighbor " << polyominoNbr.type->name << " " << polyominoNbr.offset.toString() << std::endl;
				// std::cout << "  old strength: " << oldStrengthNbr << std::endl;
				// std::cout << "  new strength: " << strengthNbr << std::endl;

				eventSet->removeDetachmentEvent(polyominoNbr, oldStrengthNbr);
				eventSet->addDetachmentEvent(polyominoNbr, strengthNbr);
			} else {
				if (overlappedMap[nbr] == 0 and
				    (minBinding == 0 or oldStrengthNbr < minBinding) and
				    strengthNbr >= minBinding) {
					// if (polyominoNbr.type->name == "h0")
					// 	std::cout << "adding attachment event for polyomino " << polyominoNbr.type->name << " " << polyominoNbr.offset.toString() << std::endl;
					eventSet->addAttachmentEvent(polyominoNbr);
				}
			}
		}
	}

	// remove overlapping attachment events
	for (Shape ovr : this->sharedData->overlapLists[sType->id]) {
		ovr.offset += polyomino.offset;

		if (ovr.offset < dimRestriction.min or ovr.offset >= dimRestriction.max)
			continue;
		
		int overlappedOvr = ++(overlappedMap[ovr]);

		for (PolyominoType *pType : ovr.type->polyominoTypes) {
			Polyomino polyominoOvr(pType, ovr.offset);
			int strengthOvr = strengthMap[polyominoOvr];
			if (strengthOvr >= minBinding and overlappedOvr == 1) {
				// std::cout << polyominoOvr.type->name << " " << polyominoOvr.offset.toString() << std::endl;
				eventSet->removeAttachmentEvent(polyominoOvr);
			}
		}
	}
}

void KineticSimulation::removePolyomino(Polyomino polyomino)
{
	int id1 = polyomino.type->id;
	// remove state from map and update neighbors
	ShapeType *sType = this->sharedData->polyominoShapes[id1];
	Shape shape(sType, polyomino.offset);
	
	if (!locationMap.count(shape)) {
		std::string errString = "attempting to remove polyomino \"" +
			polyomino.type->name + "\" from unoccupied location " +
			polyomino.offset.toString();
		throw std::runtime_error(errString);
	}

	locationMap.erase(shape);
	int strength = strengthMap[polyomino];
	eventSet->removeDetachmentEvent(polyomino, strength);

	for (Shape nbr : this->sharedData->neighborLists[sType->id]) {
		Vec3 offset = nbr.offset;
		nbr.offset += polyomino.offset;

		// ignore neighbors that don't satisfy dim restrictions
		if (nbr.offset < dimRestriction.min or nbr.offset >= dimRestriction.max)
			continue;

		Polyomino existing = polyomino;
		if (locationMap.count(nbr))
			existing = locationMap[nbr];
		
		for (PolyominoType *pType : nbr.type->polyominoTypes) {
			Polyomino polyominoNbr(pType, nbr.offset);

			int id2 = pType->id;
			int relStrength = this->sharedData->bindingMap[id1][id2][offset];

			int &strengthNbr = strengthMap[polyominoNbr];
			int oldStrengthNbr = strengthNbr;
			strengthNbr -= relStrength;

			if (polyominoNbr == existing and seedSet.count(polyominoNbr) == 0) {
				eventSet->removeDetachmentEvent(polyominoNbr, oldStrengthNbr);
				eventSet->addDetachmentEvent(polyominoNbr, strengthNbr);
			} else {
				if (overlappedMap[nbr] == 0 and
				    oldStrengthNbr >= minBinding and
				    strengthNbr < minBinding) {
					eventSet->removeAttachmentEvent(polyominoNbr);
				}
			}
		}
	}

	for (Shape ovr : this->sharedData->overlapLists[sType->id]) {
		ovr.offset += polyomino.offset;

		if (ovr.offset < dimRestriction.min or ovr.offset >= dimRestriction.max)
			continue;
		
		int overlappedOvr = --(overlappedMap[ovr]);

		for (PolyominoType *pType : ovr.type->polyominoTypes) {
			Polyomino polyominoOvr(pType, ovr.offset);
			int strengthOvr = strengthMap[polyominoOvr];
			if (strengthOvr >= minBinding and overlappedOvr == 0) {
				eventSet->addAttachmentEvent(polyominoOvr);
			}
		}
	}
}

StochasticSim::StochasticSim(std::vector<PolyominoType*> *types,
			     double Gmc, double Gse, double kf,
			     DimensionRestriction dimRestriction, int minBinding)
	: Gmc(Gmc), Gse(Gse), dimRestriction(dimRestriction), minBinding(minBinding)
{
	int maxStrength = 0;
	for (PolyominoType *type : *types) {
		// type->id = this->polyominoTypes.size();
		// this->polyominoTypes.push_back(type);

		this->polyominoTypes = types;

		int strength = 0;
		for (const Block &block : type->blocks) {
			for (const Domain &domain : block.domains) {
				strength += domain.strength;
			}
		}
		if (maxStrength < strength)
			maxStrength = strength;
	}

	this->eventSet = new StochasticEventSet(*polyominoTypes, maxStrength, Gmc, Gse, kf);
	
	// std::cout << "baking precomputation maps from type data... " << std::flush;

	// this->computeShapes();
	// this->computeBoundingBoxes();
	// this->computeNeighborLists();
	// this->computeBindingMap();
	
	// std::cout << "done" << std::endl;

	this->time = 0.0;
	this->steps = 0;
}

void StochasticSim::setPrecomputationPointers(
	std::vector<ShapeType*> *polyominoShapes,
	std::vector<ShapeType*> *shapeTypes,
	std::vector<std::vector<Shape>> *neighborLists,
	std::vector<std::vector<Shape>> *overlapLists,
	std::vector<std::vector<std::unordered_map<Vec3, int>>> *bindingMap)
{
	this->polyominoShapes = polyominoShapes;
	this->shapeTypes = shapeTypes;
	this->neighborLists = neighborLists;
	this->overlapLists = overlapLists;
	this->bindingMap = bindingMap;
}

void StochasticSim::step()
{
	StochasticEvent event = eventSet->sample();
	
	if (event.type == ATTACHMENT) {
		addPolyomino(event.polyomino);
	} else {
		removePolyomino(event.polyomino);
	}

	this->time += event.dt;
	this->steps += 1;
}

void StochasticSim::addPolyomino(const Polyomino &polyomino, bool seed)
{
	int id1 = polyomino.type->id;
	
	// update neighbors
	ShapeType *sType = (*polyominoShapes)[id1];
	Shape shape(sType, polyomino.offset);
	
	if (locationMap.count(shape)) {
		std::string errString = "attempting to add polyomino \"" +
			polyomino.type->name + "\" in occupied location " +
			polyomino.offset.toString();
		
		throw std::runtime_error(errString);
	}

	locationMap.emplace(shape, polyomino);
	int strength = strengthMap[polyomino];
	if (seed) seedSet.insert(polyomino);
	else eventSet->addDetachmentEvent(polyomino, strength);
	
	for (Shape nbr : (*neighborLists)[sType->id]) {
		Vec3 offset = nbr.offset;
		nbr.offset += polyomino.offset;

		// ignore neighbors that don't satisfy dim restrictions
		if (nbr.offset < dimRestriction.min or nbr.offset >= dimRestriction.max)
			continue;

		Polyomino existing = polyomino;
		if (locationMap.count(nbr))
			existing = locationMap[nbr];
		
		for (PolyominoType *pType : nbr.type->polyominoTypes) {
			Polyomino polyominoNbr(pType, nbr.offset);

			int id2 = pType->id;
			int relStrength = (*bindingMap)[id1][id2][offset];

			int &strengthNbr = strengthMap[polyominoNbr];
			int oldStrengthNbr = strengthNbr;
			strengthNbr += relStrength;

			if (polyominoNbr == existing and seedSet.count(polyominoNbr) == 0) {
				eventSet->removeDetachmentEvent(polyominoNbr, oldStrengthNbr);
				eventSet->addDetachmentEvent(polyominoNbr, strengthNbr);
			} else {
				if (overlappedMap[nbr] == 0 and
				    oldStrengthNbr < minBinding and
				    strengthNbr >= minBinding) {
					eventSet->addAttachmentEvent(polyominoNbr);
				}
			}
		}
	}

	// remove overlapping attachment events
	for (Shape ovr : (*overlapLists)[sType->id]) {
		ovr.offset += polyomino.offset;

		if (ovr.offset < dimRestriction.min or ovr.offset >= dimRestriction.max)
			continue;
		
		int overlappedOvr = ++(overlappedMap[ovr]);

		for (PolyominoType *pType : ovr.type->polyominoTypes) {
			Polyomino polyominoOvr(pType, ovr.offset);
			int strengthOvr = strengthMap[polyominoOvr];
			if (strengthOvr >= minBinding and overlappedOvr == 1) {
				eventSet->removeAttachmentEvent(polyominoOvr);
			}
		}
	}
}

void StochasticSim::removePolyomino(const Polyomino &polyomino)
{
	int id1 = polyomino.type->id;
	// remove state from map and update neighbors
	ShapeType *sType = (*polyominoShapes)[id1];
	Shape shape(sType, polyomino.offset);
	
	if (!locationMap.count(shape)) {
		std::string errString = "attempting to remove polyomino \"" +
			polyomino.type->name + "\" from unoccupied location " +
			polyomino.offset.toString();
		throw std::runtime_error(errString);
	}

	locationMap.erase(shape);
	int strength = strengthMap[polyomino];
	eventSet->removeDetachmentEvent(polyomino, strength);

	for (Shape nbr : (*neighborLists)[sType->id]) {
		Vec3 offset = nbr.offset;
		nbr.offset += polyomino.offset;

		// ignore neighbors that don't satisfy dim restrictions
		if (nbr.offset < dimRestriction.min or nbr.offset >= dimRestriction.max)
			continue;

		Polyomino existing = polyomino;
		if (locationMap.count(nbr))
			existing = locationMap[nbr];
		
		for (PolyominoType *pType : nbr.type->polyominoTypes) {
			Polyomino polyominoNbr(pType, nbr.offset);

			int id2 = pType->id;
			int relStrength = (*bindingMap)[id1][id2][offset];

			int &strengthNbr = strengthMap[polyominoNbr];
			int oldStrengthNbr = strengthNbr;
			strengthNbr -= relStrength;

			if (polyominoNbr == existing and seedSet.count(polyominoNbr) == 0) {
				eventSet->removeDetachmentEvent(polyominoNbr, oldStrengthNbr);
				eventSet->addDetachmentEvent(polyominoNbr, strengthNbr);
			} else {
				if (overlappedMap[nbr] == 0 and
				    oldStrengthNbr >= minBinding and
				    strengthNbr < minBinding) {
					eventSet->removeAttachmentEvent(polyominoNbr);
				}
			}
		}
	}

	for (Shape ovr : (*overlapLists)[sType->id]) {
		ovr.offset += polyomino.offset;

		if (ovr.offset < dimRestriction.min or ovr.offset >= dimRestriction.max)
			continue;
		
		int overlappedOvr = --(overlappedMap[ovr]);

		for (PolyominoType *pType : ovr.type->polyominoTypes) {
			Polyomino polyominoOvr(pType, ovr.offset);
			int strengthOvr = strengthMap[polyominoOvr];
			if (strengthOvr >= minBinding and overlappedOvr == 0) {
				eventSet->addAttachmentEvent(polyominoOvr);
			}
		}
	}
}

// bool shapeMatches(ShapeType *sType, PolyominoType *pType)
// {
// 	if (sType->coords.size() != pType->blocks.size())
// 		return false;
	
// 	for (const Block &block : pType->blocks) {
// 		bool found = false;
// 		for (const Vec3 &coord : sType->coords) {
// 			if (block.coord == coord) {
// 				found = true;
// 				break;
// 			}
// 		}
// 		if (!found)
// 			return false;
// 	}
// 	return true;
// }

// void StochasticSim::computeShapes()
// {
// 	for (PolyominoType *pType : this->polyominoTypes) {
// 		ShapeType *foundType = nullptr;
// 		for (ShapeType *sType : this->shapeTypes) {
// 			if (shapeMatches(sType, pType)) {
// 				foundType = sType;
// 				sType->polyominoTypes.push_back(pType);
// 			}
// 		}

// 		if (foundType == nullptr) {
// 			foundType = new ShapeType();
// 			foundType->polyominoTypes.push_back(pType);
// 			foundType->id = this->shapeTypes.size();
// 			for (const Block &block : pType->blocks)
// 				foundType->coords.push_back(block.coord);

// 			this->shapeTypes.push_back(foundType);
// 		}

// 		polyominoShapes.push_back(foundType);
// 	}
// }

// inline void ewMin(Vec3 &a, const Vec3 &b) {
// 	a.x = std::min(a.x, b.x);
// 	a.y = std::min(a.y, b.y);
// 	a.z = std::min(a.z, b.z);
// }

// inline void ewMax(Vec3 &a, const Vec3 &b) {
// 	a.x = std::max(a.x, b.x);
// 	a.y = std::max(a.y, b.y);
// 	a.z = std::max(a.z, b.z);
// }

// void StochasticSim::computeBoundingBoxes() {
// 	for (ShapeType *sType : this->shapeTypes) {
// 		sType->minCoord = sType->coords.at(0);
// 		sType->maxCoord = sType->coords.at(0);
		
// 		for (const Vec3 &coord : sType->coords) {
// 			ewMin(sType->minCoord, coord);
// 			ewMax(sType->maxCoord, coord);
// 		}
// 	}
// }

// inline int mh_dist(const Vec3 &a, const Vec3 &b) {
// 	return std::abs(a.x - b.x) + std::abs(a.y - b.y) + std::abs(a.z - b.z);
// }

// // returns 2 if overlaps, 1 if neighbors, 0 if neither
// int overlap(const ShapeType *s1, const Shape &s2) {
// 	int ret = 0;
// 	for (const Vec3 &coord2 : s2.type->coords) {
// 		Vec3 coord2_off = coord2 + s2.offset;
// 		for (const Vec3 &coord1 : s1->coords) {
// 			int dist = mh_dist(coord1, coord2_off);
// 			if (dist == 0)
// 				return 2;
// 			else if (dist == 1)
// 				ret = 1;
// 		}
// 	}
// 	return ret;
// }

// void StochasticSim::checkNeighbors(ShapeType *st1, ShapeType *st2)
// {
// 	Vec3 minDiff = st1->minCoord - st2->maxCoord - Vec3(1, 1, 1);
// 	Vec3 maxDiff = st1->maxCoord - st2->minCoord + Vec3(1, 1, 1);
	
// 	for (int x = minDiff.x; x <= maxDiff.x; ++x) {
// 		for (int y = minDiff.y; y <= maxDiff.y; ++y) {
// 			for (int z = minDiff.z; z <= maxDiff.z; ++z) {
// 				Shape s(st2, Vec3(x, y, z));
// 				int cmp = overlap(st1, s);
// 				if (cmp == 2) {
// 					this->overlapLists.at(st1->id).push_back(s);
// 				} else if (cmp == 1) {
// 					this->neighborLists.at(st1->id).push_back(s);
// 				}
// 			}
// 		}
// 	}
// }

// // TODO: handle dim restrictions
// void StochasticSim::computeNeighborLists()
// {
// 	this->overlapLists.resize(this->shapeTypes.size());
// 	this->neighborLists.resize(this->shapeTypes.size());
// 	for (ShapeType *st1 : this->shapeTypes) {
// 		for (ShapeType *st2 : this->shapeTypes) {
// 			this->checkNeighbors(st1, st2);
// 		}
// 	}
// }

// bool complementaryLabels(const std::string &a, const std::string &b)
// {
// 	return a + "*" == b or a == b + "*";
// }

// Vec3 dirOffset(int direction)
// {
// 	switch (direction) {
// 	case NORTH : return Vec3( 0, 1, 0);
// 	case SOUTH : return Vec3( 0,-1, 0);
// 	case EAST  : return Vec3( 1, 0, 0);
// 	case WEST  : return Vec3(-1, 0, 0);
// 	case UP    : return Vec3( 0, 0, 1);
// 	case DOWN  : return Vec3( 0, 0,-1);
// 	default    : return Vec3( 0, 0, 0);
// 	}
// }

// int getBlockStrength(const Block &b1, const Block &b2, const Vec3 &offset)
// {
// 	Vec3 diff = (b2.coord + offset) - b1.coord;
// 	if (std::abs(diff.x) + std::abs(diff.y) + std::abs(diff.z) != 1)
// 		return 0;

// 	int strength = 0;
// 	for (const Domain &d1 : b1.domains) {
// 		for (const Domain &d2 : b2.domains) {
// 			if (d1.strength != d2.strength)
// 				continue;
// 			if (d1.direction + d2.direction != 5)
// 				continue;
// 			if (!complementaryLabels(d1.label, d2.label))
// 				continue;
// 			if (diff == dirOffset(d1.direction))
// 				strength += d1.strength;
// 		}
// 	}

// 	return strength;
// }

// int getRelativeStrength(PolyominoType *pt1, PolyominoType *pt2, const Vec3 &offset)
// {
// 	int strength = 0;
// 	for (const Block &b1 : pt1->blocks) {
// 		for (const Block &b2 : pt2->blocks) {
// 			strength += getBlockStrength(b1, b2, offset);
// 		}
// 	}

// 	return strength;
// }

// void StochasticSim::computeBindingMap()
// {
// 	this->bindingMap.resize(this->polyominoTypes.size());
// 	for (unsigned i = 0; i < this->polyominoTypes.size(); ++i)
// 		this->bindingMap.at(i).resize(this->polyominoTypes.size());

// 	int strength;
// 	for (ShapeType *st1 : this->shapeTypes) {
// 		for (PolyominoType *pt1 : st1->polyominoTypes) {
// 			for (Shape s2 : this->neighborLists.at(st1->id)) {
// 				Vec3 offset = s2.offset;
// 				ShapeType *st2 = s2.type;
// 				for (PolyominoType *pt2 : st2->polyominoTypes) {
// 					strength = getRelativeStrength(pt1, pt2, offset);
// 					bindingMap.at(pt1->id).at(pt2->id)[offset] = strength;
// 				}
// 			}
// 		}
// 	}
// }

